from . import helper
from . import time_helper
from . import task_helper
from . import klass_helper
from .api_work.request import Request as ApiRequest
from .api_work.settings import ApiSetting, ApiSettingManager
from . import es_helper




import asyncio, json, time
from typing import Dict, Any
from concurrent.futures import ThreadPoolExecutor


class BaseData(object):
    def __init__(self, **kwargs):
        self += kwargs
        # 记录实例化时间，用于排序
        self.__created_at = time.time()

    def __add__(self, other: Dict):
        """ 使用加法的形式添加数据
        tmp = BaseData()
        tmp += {'a': 1}
        """
        if not isinstance(other, Dict):
            raise Exception('BaseData need dict')
        for key, val in other.items():
            self[key] = val
        return self

    def __lt__(self, other):
        """ 根据时间排序，默认升序排列 """
        return self.__created_at < other.__created_at

    def __getattribute__(self, item):
        """ 当访问的属性不存在时，返回 None """
        try:
            return super().__getattribute__(item)
        except Exception:
            return None

    def __getitem__(self, item):
        """ 该函数使实例可以像字典一样为属性赋值 """
        return getattr(self, item)

    def __setitem__(self, key, value):
        """ 该函数使实例可以像字典一样获取属性值 """
        setattr(self, key, value)

    def __delitem__(self, key):
        """ 该函数使实例可以像字典一样删除属性 """
        delattr(self, key)

    def __call__(self):
        """ 使实例可执行，返回一个属性字典 """
        ret = dict()
        for item in dir(self):
            if not item.startswith('_'):
                ret[item] = getattr(self, item)
        return ret

    def __str__(self):
        """ 序列化实例字典 """
        return json.dumps(self())


class Job:
    def __init__(self, _run=None, _callback=None, **kwargs):
        self.params = BaseData(_run=_run, _callback=_callback, **kwargs)

    def __call__(self):
        if self.params._run and asyncio.iscoroutinefunction(self.params._run):
            return asyncio.create_task(self.params._run(self.params))

    def __str__(self):
        return self.params.__str__()

    def __setitem__(self, key, value):
        self.params[key] = value

    def __getitem__(self, item):
        return self.params[item]

    def __add__(self, other):
        self.params.__add__(other)
        return self


def async_callback(f):
    def wrap_callback(self, task):
        self.tasks.append(asyncio.create_task(f(self, task)))

    return wrap_callback


class Worker(asyncio.Queue):
    def __init__(self, worker_limit=1000, queue_limit=100):
        super().__init__(worker_limit)
        self.tasks = list()
        self.run_able = True
        self.put_able = True
        self.limit_queue = asyncio.Queue(queue_limit)
        self.main_task = asyncio.create_task(self.run())

    @async_callback
    async def __default_callback(self, task):
        if task in self.tasks:
            self.tasks.remove(task)
        job: Job = task.result()
        if job and job._callback:
            if asyncio.iscoroutinefunction(job._callback):
                await job._callback(job)
            else:
                job._callback(job)
        await self.limit_queue.get()

    def shutdown_put(self):
        self.put_able = False

    async def stop(self):
        self.shutdown_put()
        while True:
            if self.empty(): break
            await asyncio.sleep(1)

        self.run_able = False
        await super().put(None)
        self.tasks.append(self.main_task)
        await asyncio.wait(self.tasks)
        self.tasks.clear()

    async def put(self, item) -> None:
        if self.put_able:
            await super(Worker, self).put(item)

    async def run(self):
        while self.run_able:
            job: Job = await self.get()
            if job:
                await self.limit_queue.put(0)
                task = job()
                if task:
                    self.tasks.append(task)
                    task.add_done_callback(self.__default_callback)


if __name__ == '__main__':
    async def do(job: Job):
        print('do job')
        return job


    async def callback(job: Job):
        print(f'callback: {job}')


    async def main():
        worker = Worker()
        for i in range(10000):
            job = Job(_run=do, _callback=callback, id=i)
            await worker.put(job)
        await worker.stop()


    asyncio.get_event_loop().run_until_complete(main())



from dfx_utils.helper import json_friendly_dumps
from typing import Dict
import time


class SuperDict:
    """
    data = SuperDict()
    data['a.b'] = 1 => {'a': {'b': 1}}
    """

    def __init__(self, _dict: dict = None, _sep='.'):
        self._sep = _sep
        self._dict = _dict or {}

    def __getitem__(self, item):
        tmp_data = None
        if self._sep in item:
            tmp_dict = self._dict
            for key in item.split(self._sep):
                tmp_data = tmp_dict.get(key)
                if tmp_data is not None:
                    if tmp_data and isinstance(tmp_data, dict):
                        tmp_dict = tmp_data
        else:
            tmp_data = self._dict.get(item)
        return SuperDict(tmp_data) if isinstance(tmp_data, dict) else tmp_data

    def __setitem__(self, key, value):
        if self._sep in key:
            tmp_dict = self._dict
            keys = key.split(self._sep)
            for idx, item in enumerate(keys):
                if idx == len(keys) - 1:
                    tmp_dict[item] = value
                else:
                    tmp_dict[item] = tmp_dict.get(item, {})
                    tmp_dict = tmp_dict[item]
        else:
            self._dict[key] = value

    def __getattribute__(self, item):
        """ 当调用的函数不存在时，去self._dict里面找 """
        try:
            return super(SuperDict, self).__getattribute__(item)
        except AttributeError:
            return self._dict.__getattribute__(item)

    def __str__(self):
        return json_friendly_dumps(self._dict)

    def clear(self):
        self._dict = {}

    def get(self, item, default=None):
        ret_data = self.__getitem__(item)
        return ret_data if ret_data is not None else default


class BaseData(object):
    def __init__(self, **kwargs):
        self += kwargs
        # 记录实例化时间，用于排序
        self._created_at = time.time()

    def __add__(self, other: Dict):
        """ 使用加法的形式添加数据
        tmp = BaseData()
        tmp += {'a': 1}
        """
        if not isinstance(other, Dict):
            raise Exception('BaseData need dict')
        for key, val in other.items():
            self[key] = val
        return self

    def __lt__(self, other):
        """ 根据时间排序，默认升序排列 """
        return self._created_at < other._created_at

    def __getattribute__(self, item):
        """ 当访问的属性不存在时，返回 None """
        try:
            return super().__getattribute__(item)
        except Exception:
            return None

    def __getitem__(self, item):
        """ 该函数使实例可以像字典一样为属性赋值 """
        return getattr(self, item)

    def __setitem__(self, key, value):
        """ 该函数使实例可以像字典一样获取属性值 """
        setattr(self, key, value)

    def __delitem__(self, key):
        """ 该函数使实例可以像字典一样删除属性 """
        delattr(self, key)

    def __call__(self):
        """ 使实例可执行，返回一个属性字典 """
        ret = dict()
        for item in dir(self):
            if not item.startswith('_'):
                ret[item] = getattr(self, item)
        return ret

    def __str__(self):
        """ 序列化实例字典 """
        return json_friendly_dumps(self())


class SuperData(BaseData):
    def __setitem__(self, key, value):
        setattr(self, key, SuperData(**value) \
            if isinstance(value, Dict) else value)

    def __call__(self):
        ret = dict()
        for item in dir(self):
            if item.startswith('_'): continue
            value = getattr(self, item)
            ret[item] = value() \
                if isinstance(value, SuperData) else value
        return ret



from dfx_utils.api_work.settings import ApiSetting


class Request:
    def __init__(self, session, setting: ApiSetting):
        """
        :session: requests.session or aiohttp.ClientSession instance
        """
        self.session = session
        self.setting = setting

    def post_sync(self):
        param, header, data = self.setting()
        return self.session.post(self.setting.request_url, params=param, headers=header, json=data, verify=False)

    async def post_async(self):
        param, header, data = self.setting()
        return await self.session.post(self.setting.request_url, params=param, headers=header, json=data, ssl=False)

    def get_sync(self):
        param, header, data = self.setting()
        return self.session.get(self.setting.request_url, params=param, headers=header, json=data, verify=False)

    async def get_async(self):
        param, header, data = self.setting()
        return await self.session.get(self.setting.request_url, params=param, headers=header, json=data, ssl=False)

    async def close(self):
        await self.session.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def __call__(self, sync=True):
        method_f = getattr(self, f'{self.setting.method}_{"sync" if sync else "async"}', None)
        if method_f:
            return method_f()
        else:
            raise Exception(f'method error {self.setting.method}')



from typing import Dict, List
from elasticsearch import Elasticsearch

import time


########### es field ##############################
class ESTypeMapping:
    Long = 'long'
    Integer = 'integer'
    Short = 'short'
    Byte = 'byte'
    Double = 'double'
    Float = 'float'
    HalfFloat = 'half_float'
    ScaledFloat = 'scaled_float'
    UnsignedLong = 'unsigned_long'
    Keyword = 'keyword'
    ConstantKeyword = 'constant_keyword'
    Wildcard = 'wildcard'
    Text = 'text'
    Date = 'date'
    IP = 'ip'
    Boolean = 'boolean'


class ESBaseFeild:
    def __init__(self, feild_name, feild_type, **properties):
        self.feild_name = feild_name
        self.properties = {'type': feild_type}
        self.properties.update(properties)

    def add_property(self, **properties):
        """ 添加字段属性 """
        self.properties.update(properties)

    def get_feild(self):
        """ 获取属性结果 """
        return {self.feild_name: self.properties}


class LongFeild(ESBaseFeild):
    """ 64 位长整型 """

    def __init__(self, feild_name, **properties):
        super(LongFeild, self).__init__(feild_name, ESTypeMapping.Long, **properties)


class IntegerFeild(ESBaseFeild):
    """ 32位整型 """

    def __init__(self, feild_name, **properties):
        super(IntegerFeild, self).__init__(feild_name, ESTypeMapping.Integer, **properties)


class ShortFeild(ESBaseFeild):
    """ 16位短整型 """

    def __init__(self, feild_name, **properties):
        super(ShortFeild, self).__init__(feild_name, ESTypeMapping.Short, **properties)


class ByteFeild(ESBaseFeild):
    """ 字节类型 """

    def __init__(self, feild_name, **properties):
        super(ByteFeild, self).__init__(feild_name, ESTypeMapping.Byte, **properties)


class DoubleFeild(ESBaseFeild):
    """ 双精度浮点数 """

    def __init__(self, feild_name, **properties):
        super(DoubleFeild, self).__init__(feild_name, ESTypeMapping.Double, **properties)


class FloatFeild(ESBaseFeild):
    """ 单精度浮点型 """

    def __init__(self, feild_name, **properties):
        super(FloatFeild, self).__init__(feild_name, ESTypeMapping.Float, **properties)


class HalfFloatFeild(ESBaseFeild):
    """ 半浮点型 """

    def __init__(self, feild_name, **properties):
        super(HalfFloatFeild, self).__init__(feild_name, ESTypeMapping.HalfFloat, **properties)


class UnsignedLongFeild(ESBaseFeild):
    """ 64位 无符号长整形 """

    def __init__(self, feild_name, **properties):
        super(UnsignedLongFeild, self).__init__(feild_name, ESTypeMapping.UnsignedLong, **properties)


class KeywordFeild(ESBaseFeild):
    """ 关键字 """

    def __init__(self, feild_name, **properties):
        super(KeywordFeild, self).__init__(feild_name, ESTypeMapping.Keyword, **properties)


class ConstantKeywordFeild(ESBaseFeild):
    """ 关键字常量 值不变 """

    def __init__(self, feild_name, **properties):
        super(ConstantKeywordFeild, self).__init__(feild_name, ESTypeMapping.ConstantKeyword, **properties)


class WildcardFeild(ESBaseFeild):
    """ 通配符关键字 完整字段搜索慢 适用于类似日志grep等操作 """

    def __init__(self, feild_name, **properties):
        super(WildcardFeild, self).__init__(feild_name, ESTypeMapping.Wildcard, **properties)


class TextFeild(ESBaseFeild):
    """ 文本字段 """

    def __init__(self, feild_name, **properties):
        super(TextFeild, self).__init__(feild_name, ESTypeMapping.Text, **properties)


class DateFeild(ESBaseFeild):
    """ 日期 """

    def __init__(self, feild_name, **properties):
        super(DateFeild, self).__init__(feild_name, ESTypeMapping.Date, **properties)


class IPFeild(ESBaseFeild):
    """ IP """

    def __init__(self, feild_name, **properties):
        super(IPFeild, self).__init__(feild_name, ESTypeMapping.IP, **properties)


class BooleanFeild(ESBaseFeild):
    """ 布尔值 """

    def __init__(self, feild_name, **properties):
        super(BooleanFeild, self).__init__(feild_name, ESTypeMapping.Boolean, **properties)


########### es field ##############################
########### base data #############################
class BaseData(object):
    def __init__(self, **kwargs):
        self += kwargs
        # 记录实例化时间，用于排序
        self._created_at = time.time()

    def __add__(self, other: Dict):
        """ 使用加法的形式添加数据
        tmp = BaseData()
        tmp += {'a': 1}
        """
        if not isinstance(other, Dict):
            raise Exception('BaseData need dict')
        for key, val in other.items():
            self[key] = val
        return self

    def __lt__(self, other):
        """ 根据时间排序，默认升序排列 """
        return self._created_at < other._created_at

    def __getattribute__(self, item):
        """ 当访问的属性不存在时，返回 None """
        try:
            return super().__getattribute__(item)
        except Exception:
            return None

    def __getitem__(self, item):
        """ 该函数使实例可以像字典一样获取属性值 """
        return getattr(self, item)

    def __setitem__(self, key, value):
        """ 该函数使实例可以像字典一样获取属性值 """
        setattr(self, key, BaseData(**value) \
            if isinstance(value, Dict) else value)

    def __delitem__(self, key):
        """ 该函数使实例可以像字典一样删除属性 """
        delattr(self, key)

    def __call__(self):
        """ 使实例可执行，返回一个属性字典 """
        ret = dict()
        for item in dir(self):
            if item.startswith('_'): continue
            value = getattr(self, item)
            ret[item] = value() \
                if isinstance(value, BaseData) else value
        return ret


########### base data #############################


class Base(object):
    _name_ = None

    def __call__(self, *args, **kwargs):
        pass


class BaseField(Base):
    def __init__(self, **kwargs):
        self += kwargs

    def __add__(self, other: Dict):
        for key, val in other.items():
            setattr(self, key, val)
        return self

    def __call__(self):
        ret = dict()
        for name in dir(self):
            if name.startswith('_'): continue
            value = getattr(self, name)
            ret[name] = value() if isinstance(value, BaseField) else value
        return ret


class Fields(object):
    def __init__(self, **kwargs):
        self.fields = list()
        self += kwargs

    def __iter__(self):
        for item in self.fields:
            yield item

    def __add__(self, other: Dict):
        for key, val in other.items():
            self.fields.append(BaseField(**{key: val}))
        return self


class BaseItem(Base):
    def __init__(self, fields: Fields):
        self.fields = fields

    def __iter__(self):
        for item in self.fields:
            yield {self._name_: item()}

    def __call__(self):
        return [item for item in self]

    def add(self, **kwargs):
        self.fields += kwargs


class Term(BaseItem):
    _name_ = 'term'


class Match(BaseItem):
    _name_ = 'match'


class Range(BaseItem):
    _name_ = 'range'


class Wildcard(BaseItem):
    _name_ = 'wildcard'


class BaseQuery(Base):
    def __init__(self, *items):
        self.items = list(items)

    def add(self, item: BaseItem):
        self.items.append(item)

    def __call__(self):
        ret = list()
        for item in self.items:
            tmp = item()
            if isinstance(tmp, List):
                ret.extend(tmp)
            elif isinstance(tmp, Dict):
                ret.append(tmp)
        return {self._name_: ret}


class Must(BaseQuery):
    _name_ = 'must'


class MustNot(BaseQuery):
    _name_ = 'must_not'


class Filter(BaseQuery):
    _name_ = 'filter'


class Should(BaseQuery):
    _name_ = 'should'


class Bool(Base):
    def __init__(self, *queries):
        self._name_ = 'bool'
        self.queries = list(queries)

    def add(self, query: Base):
        self.queries.append(query)

    def deal_queries(self, item, ret: Dict):
        for key, val in item().items():
            if ret.get(key):
                ret[key].extend(val)
            else:
                ret[key] = val
        return ret

    def __call__(self):
        ret = dict()
        for item in self.queries:
            ret = self.deal_queries(item, ret)
        if 'should' in ret:
            ret['minimum_should_match'] = 1
        return {self._name_: ret}


class Sort(Base):
    def __init__(self, **kwargs):
        self._name_ = 'sort'
        self.sorts = list()
        for key, val in kwargs.items():
            self.sorts.append(BaseField(**{key: val}))

    def __call__(self, *args, **kwargs):
        return {self._name_: [item() for item in self.sorts]}


class Collapse(Base):
    """ 有去重的效果，但是返回的数据总量是去重前的
        通过某个字段折叠数据"""

    def __init__(self, field: str):
        """ field 去重的字段 """
        self._name_ = 'collapse'
        self.field = field

    def __call__(self):
        return {self._name_: {'field': self.field}}


class Update(Base):
    def __init__(self, **params):
        self._name_ = 'update'
        self.params = params
        self.lang = 'painless'

    def __call__(self):
        source = str()
        for key in self.params:
            source += f'{";" if source else ""}ctx._source.{key}=params.{key}'
        return {'script': {'source': source, 'params': self.params, 'lang': self.lang}}


class ESPagination(Base):
    def __init__(self, page=1, page_size=20, limit=10000):
        self.page = page
        self.limit = limit
        self.page_size = page_size

    def __call__(self):
        from_ = (self.page - 1) * self.page_size
        if from_ >= self.limit:
            return {'size': 0, 'from': self.limit}

        if self.page * self.page_size > self.limit:
            self.page_size = self.limit - from_

        return {'size': self.page_size, 'from': from_}


class Q(Base):
    _name_ = 'query'
    Q_ITEM_TYPE = {'term': Term, 'match': Match, 'range': Range}
    Q_QUERY_TYPE = {'must': Must, 'filter': Filter, 'should': Should}

    def __init__(self, *queries):
        self.queries = list(queries)

    def __add__(self, other):
        return self.__and__(other)

    def __and__(self, other):
        if isinstance(other, self.__class__):
            self.queries.extend(other.queries)
        elif isinstance(other, BaseQuery):
            self.queries.append(other)
        return self

    def __or__(self, other):
        if isinstance(other, BaseQuery):
            self.queries = [Should(Bool(*self.queries), Bool(other))]
        elif isinstance(other, self.__class__):
            self.queries = [Should(Bool(*self.queries), Bool(*other.queries))]
        return self

    def __call__(self,
                 sort: Sort = None,
                 pagination: ESPagination = None,
                 collapse: Collapse = None,
                 updater: Update = None):
        ret = {self._name_: Bool(*self.queries)()}
        if sort:
            ret.update(sort())
        if pagination:
            ret.update(pagination())
        if collapse:
            # 去重
            ret.update(collapse())
        if updater:
            # 更新数据
            ret.update(updater())
        return ret

    @classmethod
    def common(cls, item_typ, query_typ, **kwargs) -> "Q":
        tmp = cls.Q_ITEM_TYPE[item_typ](Fields(**kwargs))
        return cls(cls.Q_QUERY_TYPE[query_typ](tmp))

    @classmethod
    def must(cls, item_typ, **kwargs) -> "Q":
        return cls.common(item_typ, 'must', **kwargs)

    @classmethod
    def filter(cls, item_typ, **kwargs) -> "Q":
        return cls.common(item_typ, 'filter', **kwargs)

    @classmethod
    def should(cls, item_typ, **kwargs) -> "Q":
        return cls.common(item_typ, 'should', **kwargs)


class SearchSyntax:
    """ 搜索语法 """

    def __init__(self, **commands):
        self.commands = commands

    def deal(self, key: str, sep: str = ':'):
        tmp_key, conditions = key.replace(f'{sep}', f'{sep} '), list()
        if 'AND' in tmp_key:
            conditions.append('AND')
            for item in [i.strip() for i in tmp_key.split('AND') if i]:
                if 'OR' in item:
                    sub_conditions = ['OR']
                    for item in [i.strip() for i in item.split('OR') if i]:
                        tmp = [i.strip() for i in item.split(f'{sep} ') if i]
                        sub_conditions.append({self.commands[tmp[0]]: tmp[1]})
                    conditions.append(sub_conditions)
                else:
                    tmp = [i.strip() for i in item.split(f'{sep} ') if i]
                    conditions.append({self.commands[tmp[0]]: tmp[1]})
        elif 'OR' in tmp_key:
            conditions.append('OR')
            for item in [i.strip() for i in tmp_key.split('OR') if i]:
                tmp = [i.strip() for i in item.split(f'{sep} ') if i]
                conditions.append({self.commands[tmp[0]]: tmp[1]})
        else:
            tmp = [i.strip() for i in tmp_key.split(f'{sep} ') if i]
            if len(tmp) > 1:
                conditions = {self.commands[tmp[0]]: tmp[1]}
            else:
                conditions = ['OR', {'content': key}, {'group_name': key}, {'sender': key}]
        return conditions

    def _analyze_item(self, item: Dict):
        for key, val in item.items():
            if val.startswith('-'):
                return Bool(MustNot(Match(Fields(**{key: val[1:]}))))
            else:
                return Match(Fields(**item))

    def analyze(self, conditions, sub=False):
        if isinstance(conditions, Dict):
            return Filter(self._analyze_item(conditions))

        ctrl, items = conditions[0], list()
        while conditions:
            item = conditions.pop()
            if isinstance(item, List):
                items.append(self.analyze(item, sub=True))
            elif isinstance(item, Dict):
                items.append(self._analyze_item(item))
            else:
                break
        if ctrl == 'AND':
            return Filter(*items)
        elif ctrl == 'OR':
            if sub:
                return Bool(Should(*items))
            else:
                return Should(*items)


class Result(object):
    def __init__(self, result):
        self.result = result

    def total(self):
        return int(self.result.get('hits', {}).get('total', {}).get('value', 0))

    def hits(self):
        """ 获取查询数据列表 """
        return self.result.get('hits', {}).get('hits', [])

    def __iter__(self):
        """ 遍历结果 """
        for item in self.hits():
            yield BaseData(**item['_source'])


class ESHelper(object):
    def __init__(self, es: Elasticsearch, index: str = None, index_prefix: str = None):
        self.es = es
        if index:
            self.index = index
        elif index_prefix:
            self.index_prefix = index_prefix
            self.index = self.get_index_name(index_prefix)
        else:
            self.index = None

    def flush_index(self):
        self.index = self.get_index_name(self.index_prefix)

    def get_index_name(self, prefix: str = None):
        ret = self.es.indices.get_alias().keys()
        if prefix:
            ret = [item for item in ret if item.startswith(prefix)]
        if len(ret) > 0:
            return sorted(ret, reverse=True)[0]

    def search(self, body: Dict):
        if self.index:
            return Result(self.es.search(index=self.index, body=body, request_timeout=9999))

    def insert(self, body: Dict):
        self.es.index(index=self.index, body=body, refresh=True)

    def update(self, body: Dict):
        self.es.update_by_query(index=self.index, body=body, refresh=True, request_timeout=9999)

    def exists(self, body: Dict):
        body['size'] = 0
        return self.search(body).total() > 0

    @classmethod
    def del_mapping(cls, es: Elasticsearch, index: str):
        if es.indices.exists(index):
            es.indices.delete(index)

    @classmethod
    def create_mapping(cls, es: Elasticsearch, properties: List, index: str):
        mappings = {'mappings': {'properties': dict()}}
        properties_ = mappings['mappings']['properties']
        for item in properties:
            properties_.update(item.get_feild())
        es.indices.create(index=index, body=mappings)



from dfx_utils.klass_helper import BaseData, SuperDict


class ApiSetting(SuperDict):
    def __init__(self, api_name: str, request_url: str, method: str, *args, **kwargs):
        super(ApiSetting, self).__init__(*args, **kwargs)
        self.data = BaseData()
        self.method = method.lower()
        self.api_name = api_name
        self.request_url = request_url

    def __getitem__(self, item):
        key = f'{self.api_name}.{item}'
        return super(ApiSetting, self).__getitem__(key)

    def __setitem__(self, key, value):
        key = f'{self.api_name}.{key}'
        return super(ApiSetting, self).__setitem__(key, value)

    def add(self, **kwargs):
        """ 添加数据 """
        self.data.add(**kwargs)

    def reset(self):
        """ 清除已有数据 """
        self.data = BaseData()

    def data_rule(self, key, rule, required=True):
        """ 添加数据规则 """
        tmp_key = f'data.{key}'
        self[tmp_key] = (rule, required)

    def __add_rule(self, typ, args: list):
        self[typ] = args

    def header_rule(self, *args):
        """ 添加 header 规则 """
        self.__add_rule('header', args)

    def param_rule(self, *args):
        """ 添加 param 规则 """
        self.__add_rule('param', args)

    def __check_data(self, data, setting: SuperDict = None, ret_data: dict = dict()):
        """ 检查 post data """
        tmp_data = ret_data
        for key, val in setting.items():
            if isinstance(val, dict):
                tmp_data[key] = tmp_data.get(key, {})
                self.__check_data(data, setting=val, ret_data=tmp_data[key])
            else:
                if not getattr(data, key) and val[-1]:
                    raise Exception(f'{self.api_name}: data({key}) not found')
                elif getattr(data, key).__class__.__name__ not in val[0] and val[-1]:
                    raise Exception(f'{self.api_name}: data({key}) need {val[0]}, found {type(getattr(data, key))}')
                tmp_data[key] = getattr(data, key)
        return ret_data

    def __check(self, typ, data: BaseData):
        """ 检查 param 和 header """
        ret_data = BaseData()
        for item in self.get(typ, []):
            value = getattr(data, item)
            if value is None:
                raise Exception(f'{self.api_name}: {typ}({item}) not found')
            ret_data.add(**{item: value})
        return ret_data()

    def __check_param(self, data: BaseData):
        """ 检查 param """
        param_key = f'{self.api_name}.param'
        return self.__check(param_key, data)

    def __check_header(self, data: BaseData):
        """ 检查 header """
        header_key = f'{self.api_name}.header'
        return self.__check(header_key, data)

    def __call__(self):
        """ 检查并返回 param/header/data """
        headers = self.__check_header(self.data)
        param = self.__check_param(self.data)
        key = f'{self.api_name}.data'
        ret_data = self.__check_data(self.data, self.get(key, {}))
        ret_data = {k: v for k, v in ret_data.items() if v is not None}
        return param, headers, ret_data


class ApiSettingManager(dict):
    def __init__(self, base_url: str = None, name: str = None):
        self.name = name
        self.base_url = base_url

    def __call__(self, api_name, path, method) -> ApiSetting:
        """ 创建并记录 api setting """
        request_url = f'{self.base_url}{path}'
        key = f'{self.name}.{api_name}'
        self[key] = ApiSetting(key, request_url, method)
        return self[api_name]

    def __getitem__(self, item):
        key = f'{self.name}.{item}'
        return super(ApiSettingManager, self).__getitem__(key)



from typing import Any
import asyncio, logging
from functools import partial
from datetime import datetime
import random, string, hashlib, json, datetime, time


def singleton(cls):
    """ 单例模式 """
    instances = dict()

    def inner(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return inner


def random_str(length: int) -> str:
    seed = string.digits + string.ascii_letters
    return ''.join(random.choices(seed, k=length))


def md5(_str: str):
    return hashlib.md5(_str.encode()).hexdigest()


def json_encoder(obj: Any):
    """ JSON 序列化, 修复时间 """
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S')
    return super().default(obj)


def json_decoder(obj: Any):
    """ JSON 反序列化，加载时间 """
    ret = obj
    if isinstance(obj, list):
        obj = enumerate(obj)
    elif isinstance(obj, dict):
        obj = obj.items()
    else:
        return obj

    for key, item in obj:
        if isinstance(item, (list, dict)):
            ret[key] = json_decoder(item)
        elif isinstance(item, str):
            try:
                ret[key] = datetime.datetime.strptime(item, '%Y-%m-%d %H:%M:%S')
            except ValueError:
                ret[key] = item
        else:
            ret[key] = item
    return ret


def json_friendly_loads(obj: Any):
    return json.loads(obj, object_hook=json_decoder)


def json_friendly_dumps(obj: Any, **kwargs):
    return json.dumps(obj, ensure_ascii=False, default=json_encoder, **kwargs)


def retry(f, times=2, interval=0):
    """
    from functools import partial
    f = partial(test, *args, **kwargs)
    """
    try:
        return f()
    except Exception as e:
        if times > 0:
            if interval > 0:
                time.sleep(interval)
            return retry(f, times=times - 1, interval=interval)
        raise e


async def async_retry(f, times=2, interval=0):
    """
    from functools import partial
    f = partial(test, *args, **kwargs)
    """
    try:
        return await f()
    except Exception as e:
        if times > 0:
            if interval > 0:
                await asyncio.sleep(interval)
            return await async_retry(f, times=times - 1, interval=interval)
        raise e


def asynio_loop():
    loop = asyncio.get_event_loop()
    if not loop:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop=loop)
    return loop


async def to_async(f, *args, **kwargs):
    try:
        return await asynio_loop().run_in_executor(None, partial(f, *args, **kwargs))
    except Exception as e:
        logging.exception('to_async:' + str(datetime.now()), exc_info=e)



import asyncio, logging
from datetime import datetime
from typing import List, Dict
from dfx_utils.klass_helper import BaseData
from dfx_utils.helper import random_str, json_friendly_loads


class TaskSetting:
    def __init__(self, setting: Dict = None, handlers: Dict = None):
        """
        :setting: {'redis': aio_redis_instance, 'task_tag': redis_task_tag}
            :redis_task_tag: default_value -> 'default'
        :handlers: {'handle_demo': demo}
            :demo: demo is function to handle job
        """
        self.__container = BaseData()
        if setting:
            self.__container.add(**setting)
        if handlers:
            self.__container.add(**handlers)

    def add(self, **kwargs):
        self.__container.add(**kwargs)

    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except:
            return self.__container.__getattribute__(item)

    def __call__(self):
        return self.__container()

    def __delitem__(self, key):
        self.__container.add(**{key: None})

    def __setitem__(self, key, value):
        self.__container.add(**{key: value})

    def __getitem__(self, item):
        return self().get(item)

    def get(self, item, default=None):
        return self.__getitem__(item) or default


class WorkerQueue(asyncio.Queue):
    def __init__(self, queues: List[str], setting: TaskSetting):
        super().__init__()
        self.setting, self.running = setting, True
        self.queues, self.tasks, self.task_futures = dict(), list(), list()
        # 全局并发限制 默认并发100
        self.task_limit_queue = asyncio.Queue(self.setting.runtime_limit or 100)
        for name in queues:
            self.queues[name] = asyncio.Queue()
            self.task_futures.append(asyncio.ensure_future(self.__deal_job(name)))
        self.task_futures.append(asyncio.ensure_future(self.__deal_queue()))

    def __task_callback(self, task):
        self.tasks.remove(task)

    async def __deal_job(self, name):
        queue = self.queues.get(name)
        handler = self.setting.get(name)

        if handler:
            while self.running and queue:
                job: BaseData = await queue.get()
                await self.task_limit_queue.put(0)
                try:
                    task = asyncio.create_task(handler(job()))
                    self.tasks.append(task)
                    task.add_done_callback(self.__task_callback)
                except Exception as e:
                    logging.exception(f'__deal_queue:{name}:' + str(datetime.now()), exc_info=e)
                await self.task_limit_queue.get()
            del self.setting[name]
        if queue:
            del self.queues[name]

    async def __deal_queue(self):
        setting_dict = self.setting()
        while self.running:
            job: BaseData = await self.get()
            if setting_dict.get(job.name) and self.queues.get(job.name):
                await self.queues[job.name].put(job)

    async def put(self, **kwargs) -> None:
        await super().put(BaseData(**kwargs))

    async def stop(self):
        self.running = False
        for key, queue in self.queues.items():
            await queue.put(BaseData(name=key))
        await self.put(name='stop')
        await asyncio.wait(self.task_futures)
        if self.tasks:
            await asyncio.wait(self.tasks)

    async def __call__(self):
        task_tag = self.setting.task_tag or 'default'
        while self.running and self.setting.redis:
            redis_job = await self.setting.redis.brpop(task_tag)
            await worker.put(**json_friendly_loads(redis_job))


if __name__ == '__main__':
    async def test(job: Dict):
        print('test', job)


    handlers = {'test': test, 'test1': test, 'test2': test}
    task_setting = TaskSetting({}, handlers)
    worker = WorkerQueue(list(handlers.keys()), task_setting)


    async def main():
        count = 0
        while count < 10:
            await worker.put(name='test', data=random_str(32))
            await worker.put(name='test1', data=random_str(32))
            await worker.put(name='test2', data=random_str(32))
            await asyncio.sleep(0.1)
            count += 1
        await worker.stop()


    asyncio.get_event_loop().run_until_complete(main())
